#pragma once

#include "tak_lib_cpp.h"

////////////////////////////////////////////////////////////////////
// Compile time debug options
#ifdef _DEBUG
#define TRACE_SEEKING
#endif

/**
 * Helper class to preserve the read position in a file stream.
 * Usage: {restore_file_position_helper helper(p_file, p_abort); <do file operations here>}
 */
class restore_file_position_helper
{
private:
	service_ptr_t<file> m_file;
	abort_callback & m_abort;
	t_filesize m_saved_position;
public:
	restore_file_position_helper(service_ptr_t<file> p_file, abort_callback & p_abort) : m_file (p_file), m_abort(p_abort)
	{
		m_saved_position = m_file->get_position(m_abort);
	}

	~restore_file_position_helper()
	{
		m_file->seek(m_saved_position, m_abort);
	}
};

/**
 * Helpers class that closes and re-opens a TAK decoder instances.
 * Re-opening the TAK decoder is necessary after updating file tags,
 * because we use the tagging routines from foobar2000. The file size may 
 * change without the TAK decoder noticing, which could lead to crashes
 * in tak_deco_lib. Creating a new decoder instances for the updated file
 * avoids this issue.
 */
class reopen_tak_decoder_helper
{
	pfc::ptrholder_t<tak_lib_cpp::tak_seekable_decoder> &m_decoder;
	service_ptr_t<file> m_file;
	abort_callback & m_abort;
public:
	reopen_tak_decoder_helper(pfc::ptrholder_t<tak_lib_cpp::tak_seekable_decoder> &p_decoder, service_ptr_t<file> p_file, abort_callback & p_abort) : m_decoder(p_decoder), m_file(p_file), m_abort(p_abort)
	{
		m_decoder = 0;
	}

	~reopen_tak_decoder_helper()
	{
		if (m_decoder.is_empty())
		{
			m_file->seek(0, m_abort);
			m_decoder = new tak_lib_cpp::tak_seekable_decoder(m_file);
		}
	}
};

////////////////////////////////////////////////////////////////////
// Input service implementation

// No inheritance. Our methods get called over input framework templates. See input_singletrack_impl for descriptions of what each method does.
class input_tak
{
public:
	void open(service_ptr_t<file> p_filehint, const char * p_path, t_input_open_reason p_reason, abort_callback & p_abort);

	void get_info(file_info & p_info, abort_callback & p_abort);

	t_filestats get_file_stats(abort_callback & p_abort);

	void decode_initialize(unsigned p_flags,abort_callback & p_abort);

	bool decode_run(audio_chunk & p_chunk, abort_callback & p_abort);

	bool decode_run_raw(audio_chunk & p_chunk, mem_block_container & p_raw, abort_callback & p_abort);

	void set_logger(event_logger::ptr ptr);

	void decode_seek(double p_seconds, abort_callback & p_abort);

	bool decode_can_seek();

	bool decode_get_dynamic_info(file_info & p_out, double & p_timestamp_delta);
	bool decode_get_dynamic_info_track(file_info & p_out, double & p_timestamp_delta);

	void decode_on_idle(abort_callback & p_abort);

	void retag(const file_info & p_info, abort_callback & p_abort);
	
	static bool g_is_our_content_type(const char * p_content_type);
	static bool g_is_our_path(const char * p_path,const char * p_extension);

private:
	service_ptr_t<file> m_file;

	pfc::ptrholder_t<tak_lib_cpp::tak_seekable_decoder> m_decoder;

	t_uint32 m_sample_rate;
	t_uint32 m_bits_per_sample;
	t_uint32 m_channel_count;

	pfc::array_t<t_uint8> m_buffer;
	enum {BUFFER_SIZE_IN_SAMPLES = 1024};

	event_logger::ptr m_logger;

#ifndef NDEBUG
	bool m_run_called;
#endif
};
