#ifndef __TAKLIBXX_DEBUG_H__
#define __TAKLIBXX_DEBUG_H__

#ifndef __cplusplus
#error TAKLIB++ (debug extensions) must be compiled as C++
#endif

#include "taklibxx.h"

namespace taklibxx
{
#ifdef _MSC_VER

	class preserve_fpu_state
	{
		unsigned m_old_fpu, m_old_sse2;
	public:
		inline preserve_fpu_state()
		{
			__control87_2(0, 0, &m_old_fpu, &m_old_sse2);
		}

		inline ~preserve_fpu_state()
		{
			unsigned dummy_fpu, dummy_sse2;
			int const rv = __control87_2(0, 0, &dummy_fpu, &dummy_sse2);
			if (dummy_fpu != m_old_fpu || dummy_sse2 != m_old_sse2)
			{
				pfc::dynamic_assert(dummy_fpu == m_old_fpu, "FPU state altered");
				pfc::dynamic_assert(dummy_sse2 != m_old_sse2, "SSE2 state altered");
			}
			_mm_empty();
		}
	};

	template <class T>
	class tak_ssd_ptr_preserve_fpu_flags : protected T
	{
	public:
		bool create_from_file(const char * p_path, const TtakSSDOptions &p_options, TSSDDamageCallback p_callback, void * p_callback_data)
		{
			preserve_fpu_state fpu_flags;
			return T::create_from_file(p_path, p_options, p_callback, p_callback_data);
		}

		bool create_from_stream(TtakStreamIoInterface * p_stream, void * p_stream_data, const TtakSSDOptions &p_options, TSSDDamageCallback p_callback, void * p_callback_data)
		{
			preserve_fpu_state fpu_flags;
			return T::create_from_stream(p_stream, p_stream_data, p_options, p_callback, p_callback_data);
		}

		TtakResult get_stream_info(Ttak_str_StreamInfo & p_info)
		{
			preserve_fpu_state fpu_flags;
			return T::get_stream_info(p_info);
		}

		TtakResult get_stream_encoder_info(Ttak_str_MetaEncoderInfo &p_encoder_info)
		{
			preserve_fpu_state fpu_flags;
			return T::get_stream_encoder_info(p_encoder_info);
		}

		TtakResult seek(TtakInt64 p_sample_pos)
		{
			preserve_fpu_state fpu_flags;
			return T::seek(p_sample_pos);
		}

		TtakResult read_audio(void * p_buffer, TtakInt32 p_samples, TtakInt32 & p_samples_read)
		{
			preserve_fpu_state fpu_flags;
			return T::read_audio(p_buffer, p_samples, p_samples_read);
		}

		TtakInt32 get_last_frame_bitrate()
		{
			preserve_fpu_state fpu_flags;
			return T::get_last_frame_bitrate();
		}

		TtakAPEv2Tag get_tag()
		{
			preserve_fpu_state fpu_flags;
			return T::get_tag();
		}

		using T::is_empty;
		using T::is_decoder_valid;
	};

	template <class T>
	class tak_ssd_ptr_observe_fpu_state : protected T
	{
	private:
		unsigned m_fpu_state;

		void print_state(pfc::string_base & p_out, unsigned p_state)
		{
			switch (p_state & _MCW_DN)
			{
			case _DN_SAVE:
				p_out << "_DN_SAVE";
				break;
			case _DN_FLUSH:
				p_out << "_DN_FLUSH";
				break;
			case _DN_FLUSH_OPERANDS_SAVE_RESULTS:
				p_out << "_DN_FLUSH_OPERANDS_SAVE_RESULTS";
				break;
			case _DN_SAVE_OPERANDS_FLUSH_RESULTS:
				p_out << "_DN_SAVE_OPERANDS_FLUSH_RESULTS";
				break;
			}
			p_out << " | ";
			switch (p_state & _MCW_IC)
			{
			case _IC_PROJECTIVE:
				p_out << "_IC_PROJECTIVE";
				break;
			case _IC_AFFINE:
				p_out << "_IC_AFFINE";
				break;
			}
			p_out << " | ";
			switch (p_state & _MCW_PC)
			{
			case _PC_24:
				p_out << "_PC_24";
				break;
			case _PC_53:
				p_out << "_PC_53";
				break;
			case _PC_64:
				p_out << "_PC_64";
				break;
			}
			p_out << " | ";
			switch (p_state & _MCW_RC)
			{
			case _RC_NEAR:
				p_out << "_RC_NEAR";
				break;
			case _RC_DOWN:
				p_out << "_RC_DOWN";
				break;
			case _RC_UP:
				p_out << "_RC_UP";
				break;
			case _RC_CHOP:
				p_out << "_RC_CHOP";
				break;
			}
			unsigned res = p_state & ~(_MCW_DN|_MCW_IC|_MCW_PC|_MCW_RC);
			if (res != 0)
			{
				p_out << " | 0x" << pfc::format_hex(res);
			}
		}

		inline void save_state()
		{
			_controlfp_s(&m_fpu_state, 0, 0);
			console::formatter msg;
			msg << "FPU state: ";
			print_state(msg, m_fpu_state);
		}

		inline void compare_state()
		{
			unsigned fpu_state;
			_controlfp_s(&fpu_state, 0, 0);
			if (fpu_state != m_fpu_state)
			{
				console::formatter msg;
				msg << "FPU state changed to: ";
				print_state(msg, fpu_state);
				m_fpu_state = fpu_state;
			}
		}

	public:
		bool create_from_file(const char * p_path, const TtakSSDOptions &p_options, TSSDDamageCallback p_callback, void * p_callback_data)
		{
			save_state();
			return T::create_from_file(p_path, p_options, p_callback, p_callback_data);
		}

		bool create_from_stream(TtakStreamIoInterface * p_stream, void * p_stream_data, const TtakSSDOptions &p_options, TSSDDamageCallback p_callback, void * p_callback_data)
		{
			save_state();
			return T::create_from_stream(p_stream, p_stream_data, p_options, p_callback, p_callback_data);
		}

		TtakResult get_stream_info(Ttak_str_StreamInfo & p_info)
		{
			compare_state();
			return T::get_stream_info(p_info);
		}

		TtakResult get_stream_encoder_info(Ttak_str_MetaEncoderInfo &p_encoder_info)
		{
			compare_state();
			return T::get_stream_encoder_info(p_encoder_info);
		}

		TtakResult seek(TtakInt64 p_sample_pos)
		{
			compare_state();
			return T::seek(p_sample_pos);
		}

		TtakResult read_audio(void * p_buffer, TtakInt32 p_samples, TtakInt32 & p_samples_read)
		{
			compare_state();
			return T::read_audio(p_buffer, p_samples, p_samples_read);
		}

		TtakInt32 get_last_frame_bitrate()
		{
			compare_state();
			return T::get_last_frame_bitrate();
		}

		TtakAPEv2Tag get_tag()
		{
			compare_state();
			return T::get_tag();
		}

		using T::is_empty;
		using T::is_decoder_valid;
	};
#endif

	template <class T>
	class tak_ssd_ptr_audio_md5 : public T
	{
	private:
		hasher_md5_state m_state;
		bool m_state_valid;

		static_api_ptr_t<hasher_md5> m_hasher;

	public:
		bool create_from_file(const char * p_path, const TtakSSDOptions &p_options, TSSDDamageCallback p_callback, void * p_callback_data)
		{
			m_hasher->initialize(m_state);
			m_state_valid = true;
			return T::create_from_file(p_path, p_options, p_callback, p_callback_data);
		}

		bool create_from_stream(TtakStreamIoInterface * p_stream, void * p_stream_data, const TtakSSDOptions &p_options, TSSDDamageCallback p_callback, void * p_callback_data)
		{
			m_hasher->initialize(m_state);
			m_state_valid = true;
			return T::create_from_stream(p_stream, p_stream_data, p_options, p_callback, p_callback_data);
		}

		TtakResult seek(TtakInt64 p_sample_pos)
		{
			m_state_valid = false;
			return T::seek(p_sample_pos);
		}

		TtakResult read_audio(void * p_buffer, TtakInt32 p_samples, TtakInt32 & p_samples_read)
		{
			TtakResult const rv = T::read_audio(p_buffer, p_samples, p_samples_read);
			if (m_state_valid)
			{
				if ((rv == tak_res_Ok) || (rv == tak_res_ssd_FrameDamaged))
				{
					//m_hasher->process(m_state, p_buffer, p_samples_read);
				}
				else
				{
					m_state_valid = false;
				}
			}
			return rv;
		}
	};

} // namespace taklibxx

#endif
